"""
Backwards compatible constants bridge.

For more details on this platform, please refer to the documentation
at https://home-assistant.io/components/fan.zha/
"""
# pylint: disable=W0614,W0401
from .core.const import *  # noqa: F401,F403
